from pathlib import Path

import request_models
from fastapi import APIRouter

from homecloud import homecloud_logging

root = Path(__file__).parent

router = APIRouter()
logger = homecloud_logging.get_logger("$app_name_server")


@router.get("/homecloud")
def homecloud(request: request_models.Request) -> dict:
    # You can add to the payload here if you want
    # but don't remove anything or the server will be
    # undiscoverable by homecloud clients.
    logger.info(f"{request.host} says hello")
    return {"app_name": "$app_name", "host": request.host}



import json
import socket
import time
from typing import Any

import requests
import tomlkit

from homecloud import homecloud_logging, homecloud_utils


def on_fail(func):
    """If contacting the server fails,
    keep scanning for the server and retrying
    the request."""

    def inner(self, *args, **kwargs):
        counter = 0
        while True:
            try:
                output = func(self, *args, **kwargs)
                break
            except Exception as e:
                print("Error contacting server")
                print(str(e))
                print(f"Retrying in {counter}s")
                time.sleep(counter)
                if counter < 60:
                    counter += 1
                # After three consecutive fails,
                # start scanning for the server running
                # on a different ip and/or port
                if counter > 2:
                    self.server_url = self.wheres_my_server()
        if self.send_logs and (
            len(self.log_stream.getvalue().splitlines()) >= self.log_send_thresh
        ):
            self.push_logs()
        return output

    return inner


class HomeCloudClient:
    def __init__(
        self,
        app_name: str,
        send_logs: bool = True,
        log_send_thresh: int = 10,
        log_level: str = "INFO",
        timeout: int = 10,
    ):
        """Initialize client object.

        :param app_name: The app name to use.

        :param send_logs: Whether to send logs to the server
        in addition to local logging.

        :param log_send_thresh: The number of logging events required
        before sending logs to the server and flushing the current stream.

        :param log_level: The level of events to log.

        :param timeout: Number of seconds to wait for a response
        when sending a request."""
        self.app_name = app_name
        self.host_name = socket.gethostname()
        self.send_logs = send_logs
        self.log_send_thresh = log_send_thresh
        self.timeout = timeout
        if send_logs:
            self.logger, self.log_stream = homecloud_logging.get_client_logger(
                f"{self.app_name}_client", self.host_name, log_level
            )
        else:
            self.logger = homecloud_logging.get_logger(
                f"{self.app_name}_client", log_level
            )
        self.server_url = self.check_last_server()
        if not self.server_url:
            self.server_url = self.wheres_my_server()
        self.base_payload = self.get_base_payload()

    def wheres_my_server(self) -> str:
        """Returns the server url for this app.
        Raises an exception if it can't be found."""
        try:
            message = f"Searching for {self.app_name} server."
            print(message)
            self.logger.info(message)
            server_ip, server_port = homecloud_utils.get_homecloud_servers()[
                self.app_name
            ]
            message = f"Found {self.app_name} server at {server_ip}:{server_port}"
            print(message)
            self.logger.info(message)
            self.save_server(server_ip, server_port)
        except Exception as e:
            server_ip = ""
            server_port = ""
            message = f"Failed to find {self.app_name} server."
            print(message)
            self.logger.exception(message)
        return f"http://{server_ip}:{server_port}"

    def save_server(self, ip: str, port: int):
        """Save server ip and port to homecloud_config.toml
        as 'last_server_ip' and 'last_server_port' fields."""
        config = homecloud_utils.load_config()
        config["last_server_ip"] = ip
        config["last_server_port"] = port
        homecloud_utils.save_config(config)
        self.logger.info(
            f"Saved '{ip}' and '{port}' to 'last_server_ip' and 'last_server_port' fields in 'homecloud_config.toml'."
        )

    def check_last_server(self) -> str | None:
        """Load ./homecloud_config.toml
        and see if the last known homecloud server
        for this app is active at the address.
        If it is, return the server's url.
        If not, return None."""
        self.logger.info(
            f"Checking if 'homecloud_config.toml' contains previous server information."
        )
        config = homecloud_utils.load_config()
        if not config:
            self.logger.info(f"No 'homecloud_config.toml' found.")
            return None
        if "last_server_ip" in config and "last_server_port" in config:
            ip = config["last_server_ip"]
            port = config["last_server_port"]
            self.logger.info(
                f"Checking previously contacted server at http://{ip}:{port}"
            )
        else:
            self.logger.info(f"No previous server information found.")
            return None
        if homecloud_utils.is_homecloud_server(ip, port) == self.app_name:
            self.logger.info(f"Previous server at http://{ip}:{port} is active.")
            return f"http://{ip}:{port}"
        else:
            self.logger.info(f"Previous server at http://{ip}:{port} is not active.")
            return None

    def get_base_payload(self) -> dict:
        """Can be overridden without having to override self.__init__()"""
        return {"host": self.host_name}

    def send_request(
        self, method: str, resource: str, data: dict = {}, params: dict = {}
    ) -> requests.Response:
        """Send a request to the server.

        :param method: The method to use (get, post, etc.).

        :param resource: The path location of the requested resource
        (e.g. /users/me).

        :param data: The request body.

        :param params: Url parameters."""
        data |= self.base_payload
        url = f"{self.server_url}{resource}"
        data = json.dumps(data)
        return requests.request(
            method, url, data=data, params=params, timeout=self.timeout
        )

    def push_logs(self):
        """Push log stream to the server."""
        self.logger.info(f"Pushing log stream to {self.app_name} server.")
        self.send_request(
            "post", "/clientlogs", data={"log_stream": self.log_stream.getvalue()}
        )
        self.log_stream.truncate(0)
        self.log_stream.seek(0)



import json

import requests

from homecloud import HomeCloudClient, on_fail


class app_nameClient(HomeCloudClient):
    def __init__(
        self,
        app_name: str = "$app_name",
        send_logs: bool = True,
        log_send_thresh: int = 10,
        log_level: str = "INFO",
        timeout: int = 10,
    ):
        super().__init__(app_name, send_logs, log_send_thresh, log_level, timeout)

    @on_fail
    def hello(self) -> str:
        """Contacts the server and returns the app name."""
        self.logger.debug(f"Saying hello to the {self.app_name} server.")
        return json.loads(self.send_request("get", "/homecloud").text)["app_name"]



import json
import socket
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

import requests
import tomlkit

import lanutils

root = Path(__file__).parent


def is_homecloud_server(ip: str, port: int) -> str | bool:
    """Determine if ip address and port number
    are active as a homecloud server.
    The 'homecloud' route of a homecloud server should return
    the app name the server is serving and the host
    name of the device that made the request.

    Returns the app name of the server if it
    is a homecloud server, else return False."""
    url = f"http://{ip}:{port}/homecloud"
    host = socket.gethostname()
    try:
        response = json.loads(
            requests.get(url, data=json.dumps({"host": host}), timeout=1).text
        )
        if response["host"] == host:
            return response["app_name"]
        else:
            return False
    except Exception as e:
        return False


def get_homecloud_servers(
    port_range: tuple[int, int] = (50000, 50100)
) -> dict[str, tuple[str, int]]:
    """Scan the local network for servers.

    :param port_range: The range of ports to scan for homecloud servers.
    The larger the range, the longer the scan will take.

    Returns a dictionary where the key is the app name
    and the value is a tuple containing the ip address
    and the port number serving that app.
    >>> print(get_homecloud_servers())
    >>> {"$app_name": ("10.0.0.49", 50025), "$app_name2": ("10.0.0.32", 50041)}"""
    ips = lanutils.enumerate_devices()
    # Get all open ports in port_range for all ips
    with ThreadPoolExecutor() as executor:
        threads = [executor.submit(lanutils.scan_ports, ip, port_range) for ip in ips]
    open_addresses = [
        (ip, port)
        for ip, thread in zip(ips, threads)
        for port in thread.result()
        if len(thread.result()) > 0
    ]
    # Scan ports of ips for homecloud servers
    with ThreadPoolExecutor() as executor:
        threads = [
            executor.submit(
                is_homecloud_server,
                address[0],
                address[1],
            )
            for address in open_addresses
        ]
    return {
        thread.result(): (address[0], address[1])
        for address, thread in zip(open_addresses, threads)
        if thread.result()
    }


def load_config(config_path: Path | str = None) -> dict | None:
    """Load and return toml config file.

    :param config_path: file path to the config file.
    If not given, the current working directory will be searched
    for a 'homecloud_config.toml' file."""
    if not config_path:
        config_path = Path.cwd() / "homecloud_config.toml"
    else:
        config_path = Path(config_path)
    if not config_path.exists():
        RuntimeWarning(f"No '{config_path.stem}' found in {config_path.parent}")
        return None
    return tomlkit.loads(config_path.read_text())


def save_config(config: dict, config_path: Path | str = None):
    """Save toml config file.

    :param config_path: file path to the config file.
    If not given, a file named 'homecloud_config.toml'
    will be created in the current working directory."""
    if not config_path:
        config_path = Path.cwd() / "homecloud_config.toml"
    else:
        config_path = Path(config_path)
    config_path.write_text(tomlkit.dumps(config))



from pathlib import Path

import request_models
from fastapi import APIRouter

from homecloud import homecloud_logging

root = Path(__file__).parent

router = APIRouter()
logger = homecloud_logging.get_logger("$app_name_server")


@router.post("/clientlogs")
def writelogs(request: request_models.LogsRequest):
    with (root / "logs" / f"{request.host}.log").open("a") as logfile:
        logfile.write(request.log_stream)



import argparse
import importlib
import importlib.util
import inspect
import sys
from pathlib import Path
from typing import Any

import black

root = Path(__file__).parent

""" Generate client functions from a module's functions."""


def load_module_from_file(path: Path) -> Any:
    """Load a module from a local file."""
    sys.path.insert(0, str(path.parent))
    spec = importlib.util.spec_from_file_location(path.stem, str(path))
    module = importlib.util.module_from_spec(spec)
    sys.modules[path.stem] = module
    spec.loader.exec_module(module)
    return module


def get_module(module: str | Path) -> Any:
    """Load and return a module.

    :param module: Either the name of an installed package
    or a Path object pointing to a .py file"""
    try:
        if type(module) == str:
            return importlib.import_module(module)
        else:
            return load_module_from_file(module)
    except Exception as e:
        print(e)
        print(f"Could not load {module}")


def get_functions(module) -> list[tuple]:
    """Return a list of public functions.
    Each element is a tuple containing the
    function name, value, and docstring."""
    functions = []
    for function, value in inspect.getmembers(module):
        if inspect.isclass(value):
            for class_function, class_value in inspect.getmembers(value):
                if inspect.isfunction(class_value):
                    functions.append(
                        (class_function, class_value, inspect.getdoc(class_value))
                    )
        elif inspect.isfunction(value):
            functions.append((function, value, inspect.getdoc(value)))
    origin = Path(importlib.util.find_spec(module.__name__).origin)
    # Local packages include functions from built in modules they import
    if not ("Lib" and "Python" and "site-packages" in origin.parts):
        source = inspect.getsource(module)
        functions = [
            function for function in functions if f"def {function[0]}" in source
        ]
    return [function for function in functions if not function[0].startswith("_")]


def format_definition(function: tuple[str, Any]) -> str:
    """Return a function definition string from
    a tuple containing the function name and it's value."""
    params = str(inspect.signature(function[1]))
    if "(self" not in params:
        params = params.replace("(", "(self, ", 1)
    return f"def {function[0]}{params}:"


def format_request(
    function_parameters: dict | list[str],
    method: str = None,
    resource: str = None,
) -> str:
    """Format and return the send_request line of the function.

    :param function_parameters: An iterable of parameters
    to be sent in the request body.

    :param method: Replace '$METHOD' with parameter value.

    :param resource: Replace '$RESOURCE' with parameter value."""
    data = ", ".join(
        [f'"{param}": {param}' for param in function_parameters if param != "self"]
    )
    request = f'response = self.send_request("$METHOD", "$RESOURCE", data={{{data}}})'
    if method:
        request = request.replace("$METHOD", method)
    if resource:
        request = request.replace("$RESOURCE", resource)
    return request


def format_return(return_annotation: Any) -> str:
    """Format the function return line.

    :param return_annotation: The return type(s) of the function."""
    return_ = "return "
    if return_annotation == inspect._empty:
        return_ += "response"
    elif return_annotation not in [int, float, str, bool]:
        return_ += "json.loads(response.text)"
    else:
        return_ += "response.text"
    return return_


def format_function(
    decorator: str,
    definition: str,
    doc: str,
    request: str,
    return_: str,
    indent_level: int,
    indent_ch: str,
) -> str:
    """Format function from it's parts."""
    function = ""
    indent = indent_ch * indent_level
    line = lambda s: f"{indent}{s}\n"
    for part in [decorator, definition]:
        function += line(part)
    # Indent another level for the function body
    indent += indent_ch
    for part in [doc, request, return_]:
        function += line(part)
    return function


def generate_client_functions(
    functions: list[tuple],
    default_method: str = None,
    resource_parent: str = None,
    indent_level: int = 1,
    indent_ch: str = " " * 4,
) -> str:
    """Generate client functions from a list of functions
    where each element is a tuple containing the function
    name, the function value, and the function's docstring.

    :param method: Replace '$METHOD' with parameter value.
    If not provided, function names with 'get' in them will
    replace their instance of '$METHOD' with 'get'
    and methods with no return annotation will
    have theirs replaced with 'post' .

    :param resource_parent: The resource path will be determined by the function name.
    This value will be prepended to it.
    e.g. For a module function "get_rows" and a resource_parent "database/users",
    the endpoint the client function will request is "/database/users/get-rows" .

    :param indent_level: The indent level of the function.
    If using the generated client file, this should be left as default.

    :param indent_ch: The indentation character to use.
    Default is four spaces."""
    client_functions = []

    for function in functions:
        # Reset method for each iteration
        method = default_method
        sig = inspect.signature(function[1])
        decorator = "@on_fail"
        definition = format_definition(function)
        # If no return annotation, make return type request.Response
        if sig.return_annotation == inspect._empty:
            # Reverse the definition so only the last ':' occurence is replaced
            # then reverse it back.
            definition = definition[::-1].replace(":", "->requests.Response:"[::-1], 1)[
                ::-1
            ]
        doc = f'""" {function[2] if not None else "No docstring found."} """'
        resource = f"/{function[0].replace('_','-')}"
        if resource_parent:
            resource = f"/{resource_parent}{resource}"
        if not method:
            if "get" in definition.lower():
                method = "get"
            elif sig.return_annotation == inspect._empty:
                method = "post"
        request = format_request(
            sig.parameters,
            method,
            resource,
        )
        return_ = format_return(sig.return_annotation)
        client_functions.append(
            format_function(
                decorator, definition, doc, request, return_, indent_level, indent_ch
            )
        )
    return client_functions


def append_to_client(client_path: Path, client_functions: list[str]):
    """Append client_functions to client file if a function
    with the same name doesn't already exist."""
    content = client_path.read_text()
    content += "\n" + "\n".join(
        [
            function
            for function in client_functions
            if function[: function.find("(")].strip() not in content
        ]
    )
    try:
        client_path.write_text(black.format_str(content, mode=black.Mode()))
    except Exception as e:
        print(e)
        client_path.write_text(content)


def get_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "module",
        type=str,
        help=""" The name of the installed module or a path to .py file 
        to generate functions for. """,
    )

    parser.add_argument(
        "-m",
        "--method",
        type=str,
        default=None,
        help=""" Replace method placeholders in generated functions
        with this string. If None, any function with 'get' in the name
        will have their method placeholder replaced with 'get'.""",
    )

    parser.add_argument(
        "-r",
        "--resource_parent",
        type=str,
        default=None,
        help="""The resource path will be determined by the function name.
        This value will be prepended to it.
        e.g. For a module function "get_rows" and a resource_parent "database/users",
        the endpoint the client function will request is "/database/users/get-rows" .""",
    )

    parser.add_argument(
        "-il",
        "--indent_level",
        type=int,
        default=1,
        help=""" The indent level to use for the generated
        functions when writing them to the client file.""",
    )

    parser.add_argument(
        "-ic",
        "--indent_ch",
        type=str,
        default=" " * 4,
        help=""" The character to use for indentation.
        Default is four spaces.""",
    )

    parser.add_argument(
        "-c",
        "--client",
        type=str,
        default=None,
        help=""" The path to a homecloud generated client file to append
        the generated functions to.
        If None, the current working directory will be scanned for one.""",
    )

    parser.add_argument(
        "-f",
        "--functions",
        type=str,
        default=None,
        nargs="*",
        help=""" A list of functions to add.
        If None, every function in the module file
        will be added.""",
    )

    args = parser.parse_args()
    if not args.client:
        try:
            args.client = [
                file
                for file in Path.cwd().glob("*_client.py")
                if "HomeCloudClient" in file.read_text()
            ][0]
        except Exception as e:
            raise FileNotFoundError(
                "Could not find an appropriate client file in the current directory."
            )
    if args.module.strip().endswith(".py"):
        args.module = Path(args.module)
        if not args.module.is_absolute():
            args.module = Path.cwd() / args.module

    return args


def main(args: argparse.Namespace = None):
    if not args:
        args = get_args()
    module = get_module(args.module)
    module_functions = get_functions(module)
    if args.functions:
        module_functions = [
            function for function in module_functions if function[0] in args.functions
        ]
    client_functions = generate_client_functions(
        module_functions,
        args.method,
        args.resource_parent,
        args.indent_level,
        args.indent_ch,
    )
    append_to_client(args.client, client_functions)


if __name__ == "__main__":
    main(get_args())



import inspect
import logging
from io import StringIO
from pathlib import Path

root = Path(__file__).parent


def get_logger(log_name: str, log_level: str = "INFO") -> logging.Logger:
    """Get a homecloud logger.
    All logs will be written to a "logs" sub directory of this file's
    parent directory.

    :param log_name: The name of the log, e.g. 'myapp_server' or 'myapp_client'.

    :param log_level: The level for this logger to log at. Same specifications
    as the build in logging module."""
    logger = logging.getLogger(log_name)

    if not logger.hasHandlers():
        root = Path(inspect.stack()[-1].filename).parent
        (root / "logs").mkdir(exist_ok=True, parents=True)
        handler = logging.FileHandler(str(root / "logs" / f"{log_name}.log"))
        handler.setFormatter(
            logging.Formatter(
                "{levelname}|-|{asctime}|-|{message}",
                style="{",
                datefmt="%m/%d/%Y %I:%M:%S %p",
            )
        )
        handler.setLevel(log_level)
        logger.addHandler(handler)
        logger.setLevel(log_level)
    return logger


def get_client_logger(
    log_name: str, host: str, log_level: str = "INFO"
) -> tuple[logging.Logger, StringIO]:
    logger = get_logger(log_name, log_level)
    log_stream = StringIO()
    handler = logging.StreamHandler(log_stream)
    handler.setFormatter(
        logging.Formatter(
            "{levelname}|-|{host}|-|{asctime}|-|{message}",
            style="{",
            datefmt="%m/%d/%Y %I:%M:%S %p",
        )
    )
    handler.setLevel(log_level)
    logger.addHandler(handler)
    logger.setLevel(log_level)
    logger = logging.LoggerAdapter(logger, {"host": host})
    return logger, log_stream



from pathlib import Path

import request_models
from fastapi import APIRouter

from homecloud import homecloud_logging

root = Path(__file__).parent

router = APIRouter()
logger = homecloud_logging.get_logger("$app_name_server")



import os
from pathlib import Path

import tomlkit
from fastapi import FastAPI

import lanutils
from homecloud.homecloud_logging import get_logger

"$router_imports"

root = Path(__file__).parent
config = tomlkit.loads((root / "homecloud_config.toml").read_text())

app = FastAPI()
"$router_includes"


def get_port_range() -> tuple[int, int]:
    """Get port_range from 'homecloud_config.toml'.
    Need to do all this casting because tomlkit class types
    mess things up."""
    port_range = tuple(config["port_range"])
    return (int(port_range[0]), int(port_range[1]))


def get_serving_address() -> tuple[str, int]:
    print("Obtaining ip address...")
    ip = lanutils.get_myip()[0][0]
    print("Finding available port in range...")
    port = lanutils.get_available_port(ip, get_port_range())
    return (ip, port)


def start_server(uvicorn_args: list[str] = ["--reload"]):
    logger = get_logger("$app_name_server")
    ip, port = get_serving_address()
    logger.info(f"Server started: http://{ip}:{port}")
    os.system(
        f"uvicorn {Path(__file__).stem}:app {' '.join(uvicorn_args)} --host {ip} --port {port}"
    )


if __name__ == "__main__":
    start_server(config["uvicorn_args"])



import argparse
from pathlib import Path

root = Path(__file__).parent

""" Class and CLI for generating the files needed to integrate homecloud into a project. """


def get_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "app_name",
        type=str,
        nargs="?",
        default=Path.cwd().stem,
        help=""" The name of the app to generate files for.
        If a value isn't provided, the folder name
        of the current working directory will be used.""",
    )

    parser.add_argument(
        "-r",
        "--routes",
        type=str,
        nargs="*",
        default=[],
        help=""" The routes to generate files for in addition to 'get' and 'post' routes.""",
    )

    parser.add_argument(
        "-d",
        "--destination",
        type=str,
        default=Path.cwd(),
        help=""" The directory to save the generated files to.
        By default, uses the current working directory. """,
    )

    args = parser.parse_args()
    args.destination = Path(args.destination)

    return args


class HomeCloudGenerator:
    def __init__(self, app_name: str, destination: Path, routes: list[str] = []):
        """Initialize homecloud generator object.

        :param app_name: The app name to use.

        :param destination: The destination to write the generated files to.

        :param routes: List of routes to generate files for
        in addition to 'get' and 'post' routes."""
        self.app_name = app_name
        self.destination = destination
        self.routes = ["get", "post"] + routes

    def generate_routes(self):
        """Generate route files."""
        for route in self.routes:
            if route not in ["get", "post"]:
                file = root / "router_template.py"
            else:
                file = root / f"{route}_routes.py"
            content = file.read_text().replace("$app_name", self.app_name)
            content = content.replace(
                "import request_models",
                f"import {self.app_name}_request_models",
            )
            content = content.replace(
                "request: request_models", f"request: {self.app_name}_request_models"
            )
            (self.destination / f"{self.app_name}_{route}_routes.py").write_text(
                content
            )

    def generate_server(self):
        """Generate the server file."""
        content = (root / "server.py").read_text()
        routes = [
            route.stem
            for route in self.destination.glob(f"{self.app_name}_*_routes.py")
        ]
        imports = "\n".join([f"import {route}" for route in routes])
        includes = "\n".join(
            [f"app.include_router({route}.router)" for route in routes]
        )
        for sub in [
            ('"$app_name"', self.app_name),
            ("$app_name", self.app_name),
            ('"$router_imports"', imports),
            ('"$router_includes"', includes),
        ]:
            content = content.replace(sub[0], sub[1])
        (self.destination / f"{self.app_name}_server.py").write_text(content)

    def generate_client(self):
        """Generate the client file."""
        content = (root / "client_subclass.py").read_text()
        content = content.replace(
            "app_nameClient", f"{self.app_name.capitalize()}Client"
        )
        content = content.replace("$app_name", self.app_name)
        (self.destination / f"{self.app_name}_client.py").write_text(content)

    def generate_request_models(self):
        """Generate the request_models file."""
        (self.destination / f"{self.app_name}_request_models.py").write_text(
            (root / "request_models.py").read_text()
        )

    def generate_config(self):
        """Generate the homecloud_config file."""
        (self.destination / "homecloud_config.toml").write_text(
            (root / "homecloud_config.toml").read_text()
        )

    def generate_all(self):
        """Generate all files."""
        self.generate_routes()
        # generate_routes() should be called before generate_server()
        # because the latter will scan the project directory
        # for route files to determine some of the server file content
        self.generate_server()
        self.generate_client()
        self.generate_request_models()
        self.generate_config()


def main(args: argparse.Namespace = None):
    if not args:
        args = get_args()
    generator = HomeCloudGenerator(args.app_name, args.destination, args.routes)
    generator.generate_all()


if __name__ == "__main__":
    main(get_args())
